"""
    Sftp upload functionality
"""


import os
from os.path import join
from sftp_uploader.applications.env_load import settings
from sftp_uploader.applications.git_manipulation import CustomGit
from sftp_uploader.applications.logs import CustomLogger
from sftp_uploader.applications.sftp import SFTPClient


class SftpUploader:
    """
        Sftp Upload files class

        Attributes:
            __sftp (SFTPClient): sftp client for upload by sftp
            __git (CustomGit): git for getting modified files
    """

    __sftp: SFTPClient = None
    __git: CustomGit = None
    logger: CustomLogger = None

    def __init__(self, 
                 sftp_client: SFTPClient,
                 git_client: CustomGit,
                 logger: CustomLogger
                 ) -> None:
        self.logger = logger
        self.__sftp = sftp_client
        self.__git = git_client
        
    def upload_to_remote(self):
        modified_file_paths: list[str] = self.__git.get_different_files_paths_from_latest_commit()
        self.logger.info(f"Get {len(modified_file_paths)} files to upload")

        for filepath in modified_file_paths:
            self.__sftp.upload_file_to_remote(local_path=filepath,
                                              remote_path=os.path.join(settings.remote_path,
                                                                       filepath))

        self.logger.info('File update complete.')




import os
import json
from pydantic import BaseSettings

from sftp_uploader.constants.applications.env_load import CONFIG_FILE_NAME


class Settings(BaseSettings):
    host: str = None
    port: int = None

    user: str = None
    password: str = None

    remote_path: str = None

    def load_settings(self):
        """
            Load settings if exists, in another case create config file
        """
        if os.path.exists(CONFIG_FILE_NAME):
            self.__load_from_config_file()

        else:
            self.__create_config_file()

    def __load_from_config_file(self):
        """
            Load config from file
        """
        
        self.parse_file(CONFIG_FILE_NAME)

    def __get_config_data_from_user(self) -> tuple[str, str, str, str]:
        """
            Getting data from user
        """
        host = input('Host: ')
        port = input('Port (default=22, press Enter to skip): ')
        if not port:
            port = '22'
        user = input('User: ')
        password = input('Password: ')
        remote_path = input('Remote path from server (project directory, you need to create it and path to it set it here): ')

        return dict(host=host, 
                    port=port, 
                    user=user,
                    password=password,
                    remote_path=remote_path,
                    )

    def __save_config(self, 
                      auth_data: dict):
        """
            Save config file

            TODO:
                * add to gitignore configfile
        """
        with open(CONFIG_FILE_NAME,
                  'w',
                  encoding='utf-8') as config_file:
            json.dump(auth_data, config_file)

    def __create_config_file(self):
        """
            Create config and save it to file
        """
        auth_data: dict = self.__get_config_data_from_user()
        self.__save_config(auth_data=auth_data)

        print(f'Your data is successfully save to file `{CONFIG_FILE_NAME}`, if you want to change it, you can change it here')

        self.parse_file(CONFIG_FILE_NAME)

 
settings = Settings()
settings.load_settings()
settings = Settings.parse_file(CONFIG_FILE_NAME)




"""
    All git manipulation logic to git manipulation application here
"""


from dependency_injector.wiring import Provide, inject
from sftp_uploader.applications.dependency_container import DependencyContainer

from sftp_uploader.applications.git_manipulation import CustomGit
from sftp_uploader.constants.applications.env_load import CONFIG_FILE_NAME


@inject
def add_config_file_to_gitignore(git: CustomGit = Provide[DependencyContainer.Git]):
    """
        Add config file package to gitignore
    """
    git.add_to_gitignote(CONFIG_FILE_NAME)


@inject
def add_sftp_upload_in_prehook(git: CustomGit = Provide[DependencyContainer.Git]):
    """
        Add config file package to gitignore
    """
    git.add_prehook_upload()




"""
    Describe project dependencies classes
"""


from dependency_injector.containers import DeclarativeContainer
from dependency_injector.providers import Singleton 
from sftp_uploader.applications.env_load import settings
from sftp_uploader.applications.git_manipulation import CustomGit
from sftp_uploader.applications.logs import CustomLogger
from sftp_uploader.applications.sftp_uploader import SftpUploader as _SftpUploader

from sftp_uploader.applications.ssh import SSHClient as CustomSSHClient
from sftp_uploader.applications.sftp import SFTPClient as CustomSFTPClient
from sftp_uploader.constants.applications.dependency_container import GIT_CLIENT_LOGGER_NAME, SFTP_CLIENT_LOGGER_NAME, SFTP_UPLOADER_LOGGER_NAME, SSH_CLIENT_LOGGER_NAME



class DependencyContainer(DeclarativeContainer):
    """
        Contaner with all dependencies
    """
    SSHClientLogger = Singleton(CustomLogger,
                                logger_name=SSH_CLIENT_LOGGER_NAME)
    SSHClient = Singleton(CustomSSHClient,
                          host=settings.host,
                          port=settings.port,
                          user=settings.user,
                          password=settings.password,
                          logger=SSHClientLogger,
                          )

    SFTPClientLogger = Singleton(CustomLogger,
                                 logger_name=SFTP_CLIENT_LOGGER_NAME)
    SFTPClient = Singleton(CustomSFTPClient,
                           ssh_client=SSHClient,
                           remote_root_path=settings.remote_path,
                           logger=SFTPClientLogger,
                           )

    GitClientLogger = Singleton(CustomLogger,
                                logger_name=GIT_CLIENT_LOGGER_NAME,
                                )
    Git = Singleton(CustomGit,
                    logger=GitClientLogger,
                    )

    SftpUploaderLogger = Singleton(CustomLogger,
                                   logger_name=SFTP_UPLOADER_LOGGER_NAME,
                                   )
    SftpUploader = Singleton(_SftpUploader,
                             sftp_client=SFTPClient,
                             git_client=Git,
                             logger=SftpUploaderLogger
                             )





"""
    Package main sources
"""


from sftp_uploader.applications.dependency_container import DependencyContainer as __DependencyContainer
from sftp_uploader.applications.pre_commit_actions import pre_commit_actions
from sftp_uploader.services.applications import sftp_uploader as __sftp_uploader
from sftp_uploader.services.applications import git_manipulation as __git_manipulation


def init_dependencies():
    """
        Initialize dependency wiring
    """
    dependency_container = __DependencyContainer()
    dependency_container.init_resources()
    dependency_container.wire(modules=[__sftp_uploader,
                                       __git_manipulation])


init_dependencies()




"""
    Poetry post install (as package) commands here
"""


from sftp_uploader.services.applications.git_manipulation import add_config_file_to_gitignore, add_sftp_upload_in_prehook


def run():
    """
        Actions after installing package
    """
    add_config_file_to_gitignore()
    add_sftp_upload_in_prehook()
    



"""
    Create a logger for package
"""


import logging

from sftp_uploader.constants.applications.logs import LOGGING_FORMAT


class CustomLogger(logging.Logger):
    """
        Custom class for logging in other classes
        
        Attributes:
            logger (logging.Logger): object logger
    """
    def __init__(self, 
                 logger_name: str) -> None:
        super().__init__(logger_name, 
                         level=logging.INFO,
                         )
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        stream_handler.setFormatter(logging.Formatter(LOGGING_FORMAT))
        self.addHandler(stream_handler)
        



"""
    Describe what we need to do before commit is complete
"""


from pathlib import Path
import subprocess
import sys
from sftp_uploader.services.applications.sftp_uploader import update_files

from sftp_uploader.services.exceptions.pre_commit_actions import CannotActivatePythonVenv


def setup_env():
        # activate_this_script = 'venv/bin/activate_this.py'
    #
    # if Path(activate_this_script).absolute().exists():
    #     exec(open(activate_this_script).read(),
    #          {'__file__': activate_this_script})
    # else:
    #     raise CannotActivatePythonVenv

    activate_this_script = 'venv/bin/activate'

    if Path(activate_this_script).absolute().exists():
        activate_script = f"{activate_this_script}"
        command = f"source {activate_script}; {sys.executable}"
        subprocess.run(command, 
                       shell=True,
                       check=True)

    else:
        raise CannotActivatePythonVenv


def pre_commit_actions():
    setup_env()
    update_files()





GIT_IGNORE_FILENAME = '.gitignore'

GIT_HOOK_SCRIPT_TO_UPLOAD = """#!/bin/sh

handle_exception() {
    arch -arm64 python -c "import sftp_uploader; sftp_uploader.pre_commit_actions()"
}

trap handle_exception ERR

python -c "import sftp_uploader; sftp_uploader.pre_commit_actions()"

exit 0
"""



GIT_HOOKS_FOLDER = '.git/hooks/'
GIT_HOOK_PRECOMMIT_FILE = f'{GIT_HOOKS_FOLDER}pre-commit'



"""
    Work with sftp protocol with custom sftp class
"""


import os
from os.path import isdir
from typing import Callable
from sftp_uploader.applications.logs import CustomLogger
from sftp_uploader.applications.ssh import SSHClient
from paramiko import SFTPClient as ParamikoSFTPClient

from sftp_uploader.services.exceptions.sftp import ErrorUploadSFTPException


class SFTPClient:
    """
        Custom sftp client for our needs
    """
    __ssh: SSHClient = None
    __sftp: ParamikoSFTPClient = None
    logger: CustomLogger = None
    __remote_root_path: str = None

    def __init__(self,
                 ssh_client: SSHClient,
                 remote_root_path: str,
                 logger: CustomLogger,
                 ) -> None:
        self.logger = logger
        self.__ssh = ssh_client
        self.__sftp = self.__ssh.get_sftp_client()
        self.__remote_root_path = remote_root_path

    def __del__(self):
        self.__sftp.close()

    def __get_directories_in_path(self, 
                                  path: str):
        """
            Get all directories to file
        """
        directories = []

        while path not in ('', '/'):
            path, directory = os.path.split(path)
            directories.append(directory)

        return directories[::-1]       

    def __check_exist_directory_to_upload(self, 
                                          remote_path: str):
        """
            Check to exists remote directory, if not - create it
        """
        directories = self.__get_directories_in_path(os.path.dirname(remote_path))
        exists_path = ''

        while directories:
            new_path = os.path.join(exists_path, 
                                    directories.pop(0))
            new_remote_path = os.path.join(self.__remote_root_path,
                                           new_path)
            try:
                self.__sftp.stat(new_remote_path)

            except FileNotFoundError:
                self.__sftp.mkdir(new_remote_path)

            finally:
                exists_path = new_path

    def __remove_from_remote(self, 
                             local_path: str,
                             remote_path: str):
        """
            If file was deleted, remove it on remote server
        """
        remove_method: Callable = None

        try:
            self.__sftp.stat(remote_path)

        except Exception as e:
            self.logger.info(f'File `{remote_path}` already delete')

        else:
            if os.path.isdir(local_path):
                remove_method = self.__sftp.rmdir

            else:
                remove_method = self.__sftp.remove

            remove_method(remote_path)
            self.logger.info(f'Success DELETE file `{remote_path}` from remote')

    def __upload_to_remote(self,
                           local_path: str,
                           remote_path: str):
        """
            Upload file to remote
        """
        self.__sftp.put(local_path, 
                        remote_path)
        self.logger.info(f'Success UPLOAD file `{local_path}` to `{remote_path}`')

    def upload_file_to_remote(self,
                              local_path: str,
                              remote_path: str):
        """
            Upload local file to remote server
        """
        self.logger.info(f'Try to modify file `{local_path}` to `{remote_path}`')
        method_to_work_with_sftp: Callable = None

        try:
            self.__check_exist_directory_to_upload(local_path)

            if not os.path.exists(local_path):
                method_to_work_with_sftp = self.__remove_from_remote

            else:
                method_to_work_with_sftp = self.__upload_to_remote

            method_to_work_with_sftp(local_path=local_path,
                                     remote_path=remote_path)

        except Exception as e:
            self.logger.exception(f'In upload to server error - `{e}`\nlocal file - `{local_path}`\nremote file - `{remote_path}`')
            raise ErrorUploadSFTPException





"""
    Work with ssh connection throught custom SSHClient
"""


from paramiko import SSHClient as ParamikoSSHClient
from paramiko import SFTPClient as ParamikoSFTPClient
from paramiko import AutoAddPolicy
from sftp_uploader.applications.logs import CustomLogger

from sftp_uploader.constants.applications.ssh import AUTH_TIMEOUT
from sftp_uploader.services.exceptions.sftp import ErrorConnectSFTPException
from sftp_uploader.services.exceptions.ssh import ErrorConnectSSHException


class SSHClient:
    """
        Custom class for ssh connection

        Attributes:
            __ssh (paramiko.SSHClient): Main ssh client
    """
    __ssh: ParamikoSSHClient = None
    __sftp: ParamikoSFTPClient= None
    logger: CustomLogger = None

    def __init__(self,
                 host: str,
                 port: int,
                 user: str,
                 password: str,
                 logger: CustomLogger,
                 ) -> None:
        self.logger = logger
        self.__ssh = ParamikoSSHClient()
        self.__ssh.set_missing_host_key_policy(AutoAddPolicy())
        self.__connect_with_ssh(host,
                                port,
                                user,
                                password)

    def __connect_with_ssh(self,
                           host: str,
                           port: int,
                           user: str,
                           password: str,
                           ):
        """
            Function to try connect by ssh
        """
        try:
            self.__ssh.connect(hostname=host, 
                               port=port,
                               username=user,
                               password=password,
                               auth_timeout=AUTH_TIMEOUT)
        except Exception as e:
            self.logger.exception(f'Error in create connection to SSH - `{e}`')
            raise ErrorConnectSSHException

        else:
            self.logger.info('SSH connect open success')

    def __del__(self):
        self.__ssh.close()
        self.logger.info('SSH connect close success')

    def __enter__(self):
        return self

    def __exit__(*args, **kwargs):
        return

    def __open_sftp_client(self) -> ParamikoSFTPClient:
        """
            Open sftp client
        """
        try:
            self.__sftp = self.__ssh.open_sftp()

        except Exception as e:
            self.logger.exception(f'Error in create connection to SFTP - `{e}`')
            raise ErrorConnectSFTPException

        else:
            self.logger.info('SFTP connect open success')

        return self.__sftp

    def get_sftp_client(self) -> ParamikoSFTPClient:
        """
            Get sftp client if exists, in another case create and get it
        """
        return self.__sftp and self.__sftp or self.__open_sftp_client()

    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sftp_uploader',
 'sftp_uploader.applications',
 'sftp_uploader.constants.applications',
 'sftp_uploader.services',
 'sftp_uploader.services.applications',
 'sftp_uploader.services.exceptions']

package_data = \
{'': ['*']}

install_requires = \
['dependency-injector>=4.41.0,<5.0.0',
 'gitpython>=3.1.31,<4.0.0',
 'paramiko>=3.1.0,<4.0.0',
 'pydantic>=1.10.7,<2.0.0',
 'python-dotenv>=0.21.0,<0.22.0']

entry_points = \
{'console_scripts': ['sftp_setup = sftp_uploader.script_to_setup_package:run']}

setup_kwargs = {
    'name': 'sftp-uploader',
    'version': '1.0.15',
    'description': 'Package for upload data in before commit uncommited files to sftp server',
    'long_description': '# TODO\n\n[ ] Make title, description, which problem is solve, metter of this project description\n\n[ ] Add more [classifiers](https://pypi.org/classifiers/)\n\n[ ] Add tests\n\n[ ] Add additional links to project conf\n\n[ ] Add poetry scripts and pre-commit scripts\n\n[ ] Add poetry post install, if git exists andd to gitignore and setup pre commit hook, and setup config module, if not exists, create and setup it\n',
    'author': 'Moonvent',
    'author_email': 'moonvent@proton.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/moonvent/sftp_uploader/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from dependency_injector.wiring import Provide, inject

from sftp_uploader.applications.dependency_container import DependencyContainer
from sftp_uploader.applications.sftp_uploader import SftpUploader


@inject
def update_files(sftp_upload: SftpUploader = Provide[DependencyContainer.SftpUploader]):
    sftp_upload.upload_to_remote()





"""
    Work with git
"""


import os
import stat
from pathlib import Path
from git import DiffIndex, Repo

from sftp_uploader.applications.logs import CustomLogger
from sftp_uploader.constants.applications.git_manipulation import GIT_HOOK_PRECOMMIT_FILE, GIT_HOOK_SCRIPT_TO_UPLOAD, GIT_HOOKS_FOLDER, GIT_IGNORE_FILENAME


class CustomGit:
    """
        Custom git class for our manipulations

        Attributes:
            __repo (Repo): local repo path
            __gitignore (str): local path to gitignore
    """
    logger: CustomLogger = None

    __repo: Repo = None
    __gitignore: str = None

    def __init__(self,
                 logger: CustomLogger,
                 ) -> None:
        self.logger = logger
        self.__get_repo()

    def __set_repo_object(self):
        self.__repo = Repo('.')

    def __is_exist_git(self) -> bool:
        """
            Check, exists git repo or not
        """
        try:
            self.__set_repo_object()

        except Exception as e:
            self.logger.exception('Git doesn\'t exist')
            return False

        else:
            return True

    def __create_repo(self):
        """
            Create git repo in local
        """
        self.__repo = Repo.init('')

    def __get_repo(self):
        """
            Init repo var
        """
        if not self.__is_exist_git:
            self.__create_repo()

        self.__set_repo_object()

    def __get_modified_files(self) -> DiffIndex:
        """
            Get modified files

            Returns:
                DiffIndex: smth like list with modified files
        """
        head_commit = self.__repo.head.commit
        result = head_commit.diff(None)
        self.logger.info(f'Find {len(result)} files to upload')
        return result

    def __get_modified_files_paths(self, 
                                   files: DiffIndex) -> list[str]:
        """
            Get modified files paths

            Args: 
                files (DiffIndex): iterable with changed files

            Returns:
                list[str]: list with mofidied files paths
        """
        return [file.a_path for file in files]

    def get_different_files_paths_from_latest_commit(self) -> list[str]:
        """
            Get modified files

            Returns:
                DiffIndex: smth like list with modified files
        """
        modified_files = self.__get_modified_files()
        paths = self.__get_modified_files_paths(files=modified_files)
        return paths

    def __is_exist_gitignore(self) -> bool:
        return os.path.exists(GIT_IGNORE_FILENAME)

    def __create_gitignore(self):
        open(GIT_IGNORE_FILENAME, 
             'w').close()
        self.logger.info('Successfully create git repo')
        self.__repo.git.add(GIT_IGNORE_FILENAME)
        self.logger.info('Successfully add gitignore file')

    def __check_gitignore_object(self):
        if not self.__is_exist_gitignore():
            self.__create_gitignore()

    def add_to_gitignote(self, 
                         filename: str):
        """
            Add filename to gitignore if exist
        """
        self.__check_gitignore_object()

        file = Path(GIT_IGNORE_FILENAME)

        if filename not in file.read_text():
            with open(file,
                      'a') as gitignore:
                gitignore.write(f'\n{filename}\n')
            self.logger.info('Successfully add your sftp config to gitignore')

    def __check_git_hook_pre_commit(self):
        """
            Check on exists need hook file
        """
        if not Path(GIT_HOOK_PRECOMMIT_FILE).exists():
            open(GIT_HOOK_PRECOMMIT_FILE, 
                 'w').close()

            os.chmod(GIT_HOOK_PRECOMMIT_FILE,
                     stat.S_IRWXU)
            # pre-commit file must be executable
            self.logger.info('Successfully create pre-commit file')

    def __add_upload_to_sftp_bash_code(self):
        """
            Add script to start python scripts before commit
        """
        precommit_file = Path(GIT_HOOK_PRECOMMIT_FILE)

        if GIT_HOOK_SCRIPT_TO_UPLOAD not in precommit_file.read_text():
            with open(GIT_HOOK_PRECOMMIT_FILE, 
                      'a') as pre_commit_bash_file:
                pre_commit_bash_file.write(f'\n{GIT_HOOK_SCRIPT_TO_UPLOAD}')
                self.logger.info('Successfully add script on load to sftp to pre-commit file')
            

    def add_prehook_upload(self):
        """
            Add to prehooks upload to sftp server
        """
        self.__check_git_hook_pre_commit()
        self.__add_upload_to_sftp_bash_code()

